#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..GXDateTime import GXDateTime
from ..enums import ObjectType, DataType
from .enums import ChargeType
from .GXUnitCharge import GXUnitCharge
from .GXChargeTable import GXChargeTable

# pylint: disable=too-many-instance-attributes,too-few-public-methods
class GXDLMSCharge(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSCharge
    """

    def __init__(self, ln="0.0.19.20.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.CHARGE, ln, sn)
        self.totalAmountPaid = 0
        self.priority = 0
        self.unitChargeActivationTime = None
        self.period = 0
        self.chargeConfiguration = 0
        self.lastCollectionTime = None
        self.lastCollectionAmount = 0
        self.totalAmountRemaining = 0
        self.proportion = 0
        self.unitChargeActive = GXUnitCharge()
        self.unitChargePassive = GXUnitCharge()
        self.chargeType = ChargeType.CONSUMPTION_BASED_COLLECTION

    def getValues(self):
        return [self.logicalName,
                self.totalAmountPaid,
                self.chargeType,
                self.priority,
                self.unitChargeActive,
                self.unitChargePassive,
                self.unitChargeActivationTime,
                self.period,
                self.chargeConfiguration,
                self.lastCollectionTime,
                self.lastCollectionAmount,
                self.totalAmountRemaining,
                self.proportion]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  TotalAmountPaid
        if all_ or self.canRead(2):
            attributes.append(2)
        #  ChargeType
        if all_ or self.canRead(3):
            attributes.append(3)
        #  Priority
        if all_ or self.canRead(4):
            attributes.append(4)
        #  UnitChargeActive
        if all_ or self.canRead(5):
            attributes.append(5)
        #  UnitChargePassive
        if all_ or self.canRead(6):
            attributes.append(6)
        #  UnitChargeActivationTime
        if all_ or self.canRead(7):
            attributes.append(7)
        #  Period
        if all_ or self.canRead(8):
            attributes.append(8)
        #  ChargeConfiguration
        if all_ or self.canRead(9):
            attributes.append(9)
        #  LastCollectionTime
        if all_ or self.canRead(10):
            attributes.append(10)
        #  LastCollectionAmount
        if all_ or self.canRead(11):
            attributes.append(11)
        #  TotalAmountRemaining
        if all_ or self.canRead(12):
            attributes.append(12)
        #  Proportion
        if all_ or self.canRead(13):
            attributes.append(13)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 13

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 5

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.INT32
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.UINT8
        elif index == 5:
            ret = DataType.STRUCTURE
        elif index == 6:
            ret = DataType.STRUCTURE
        elif index == 7:
            ret = DataType.OCTET_STRING
        elif index == 8:
            ret = DataType.UINT32
        elif index == 9:
            ret = DataType.BITSTRING
        elif index == 10:
            ret = DataType.OCTET_STRING
        elif index == 11:
            ret = DataType.INT32
        elif index == 12:
            ret = DataType.INT32
        elif index == 13:
            ret = DataType.UINT16
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    @classmethod
    def getUnitCharge(cls, charge):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(3)
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(2)
        _GXCommon.setData(bb, DataType.INT8, charge.chargePerUnitScaling.commodityScale)
        _GXCommon.setData(bb, DataType.INT8, charge.chargePerUnitScaling.priceScale)
        bb.setUInt8(DataType.STRUCTURE)
        bb.setUInt8(3)
        if charge.commodity.target is None:
            _GXCommon.setData(bb, DataType.UINT16, 0)
            bb.setUInt8(DataType.OCTET_STRING)
            bb.setUInt8(6)
            bb.setUInt8(0)
            bb.setUInt8(0)
            bb.setUInt8(0)
            bb.setUInt8(0)
            bb.setUInt8(0)
            bb.setUInt8(0)
            _GXCommon.setData(bb, DataType.INT8, 0)
        else:
            _GXCommon.setData(bb, DataType.UINT16, charge.commodity.target.objectType)
            _GXCommon.setData(bb, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(charge.commodity.target.logicalName))
            _GXCommon.setData(bb, DataType.INT8, charge.commodity.index)
        bb.setUInt8(DataType.ARRAY)
        if charge.chargeTables is None:
            bb.setUInt8(0)
        else:
            _GXCommon.setObjectCount(len(charge.chargeTables), bb)
            for it in charge.chargeTables:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(2)
                _GXCommon.setData(bb, DataType.OCTET_STRING, it.index)
                _GXCommon.setData(bb, DataType.INT16, it.chargePerUnit)
        return bb.array()

    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.totalAmountPaid
        elif e.index == 3:
            ret = self.chargeType.value
        elif e.index == 4:
            ret = self.priority
        elif e.index == 5:
            ret = self.getUnitCharge(self.unitChargeActive)
        elif e.index == 6:
            ret = self.getUnitCharge(self.unitChargePassive)
        elif e.index == 7:
            ret = self.unitChargeActivationTime
        elif e.index == 8:
            ret = self.period
        elif e.index == 9:
            ret = self.chargeConfiguration
        elif e.index == 10:
            ret = self.lastCollectionTime
        elif e.index == 11:
            ret = self.lastCollectionAmount
        elif e.index == 12:
            ret = self.totalAmountRemaining
        elif e.index == 13:
            ret = self.proportion
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    @classmethod
    def setUnitCharge(cls, settings, charge, value):
        tmp = value
        tmp2 = tmp[0]
        charge.chargePerUnitScaling.commodityScale = tmp2[0]
        charge.chargePerUnitScaling.priceScale = tmp2[1]
        tmp2 = tmp[1]
        ot = ObjectType(tmp2[0])
        ln = _GXCommon.toLogicalName(tmp2[1])
        charge.commodity.target = settings.objects.findByLN(ot, ln)
        charge.commodity.index = tmp2[2]
        charge.chargeTables = []
        tmp2 = tmp[2]
        for tmp3 in tmp2:
            it = tmp3
            item = GXChargeTable()
            item.index = it[0]
            item.chargePerUnit = it[1]
            charge.chargeTables.append(item)

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.totalAmountPaid = e.value
        elif e.index == 3:
            self.chargeType = ChargeType(e.value)
        elif e.index == 4:
            self.priority = e.value
        elif e.index == 5:
            self.setUnitCharge(settings, self.unitChargeActive, e.value)
        elif e.index == 6:
            self.setUnitCharge(settings, self.unitChargePassive, e.value)
        elif e.index == 7:
            self.unitChargeActivationTime = _GXCommon.changeType(e.value, DataType.DATETIME)
        elif e.index == 8:
            self.period = e.value
        elif e.index == 9:
            self.chargeConfiguration = str(e.value)
        elif e.index == 10:
            self.lastCollectionTime = _GXCommon.changeType(e.value, DataType.DATETIME)
        elif e.index == 11:
            self.lastCollectionAmount = e.value
        elif e.index == 12:
            self.totalAmountRemaining = e.value
        elif e.index == 13:
            self.proportion = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @classmethod
    def loadUnitChargeActive(cls, reader, name, charge):
        pass

    def load(self, reader):
        self.totalAmountPaid = reader.readElementContentAsInt("TotalAmountPaid")
        self.chargeType = ChargeType(reader.readElementContentAsInt("ChargeType"))
        self.priority = int(reader.readElementContentAsInt("Priority"))
        self.loadUnitChargeActive(reader, "UnitChargeActive", self.unitChargeActive)
        self.loadUnitChargeActive(reader, "UnitChargePassive", self.unitChargePassive)
        tmp = reader.readElementContentAsString("UnitChargeActivationTime")
        if tmp:
            self.unitChargeActivationTime = GXDateTime(tmp)
        self.period = reader.readElementContentAsInt("Period")
        self.chargeConfiguration = reader.readElementContentAsString("ChargeConfiguration")
        tmp = reader.readElementContentAsString("LastCollectionTime")
        if tmp:
            self.lastCollectionTime = GXDateTime(tmp)
        self.lastCollectionAmount = reader.readElementContentAsInt("LastCollectionAmount")
        self.totalAmountRemaining = reader.readElementContentAsInt("TotalAmountRemaining")
        self.proportion = reader.readElementContentAsInt("Proportion")

    @classmethod
    def saveUnitChargeActive(cls, writer, name, charge):
        pass

    def save(self, writer):
        writer.writeElementString("TotalAmountPaid", self.totalAmountPaid)
        if self.chargeType:
            writer.writeElementString("ChargeType", self.chargeType.value)
        writer.writeElementString("Priority", self.priority)
        self.saveUnitChargeActive(writer, "UnitChargeActive", self.unitChargeActive)
        self.saveUnitChargeActive(writer, "UnitChargePassive", self.unitChargePassive)
        writer.writeElementString("UnitChargeActivationTime", self.unitChargeActivationTime)
        writer.writeElementString("Period", self.period)
        writer.writeElementString("ChargeConfiguration", self.chargeConfiguration)
        writer.writeElementString("LastCollectionTime", self.lastCollectionTime)
        writer.writeElementString("LastCollectionAmount", self.lastCollectionAmount)
        writer.writeElementString("TotalAmountRemaining", self.totalAmountRemaining)
        writer.writeElementString("Proportion", self.proportion)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
import pkg_resources
from .enums import Standard, ObjectType, DataType
from .GXStandardObisCodeCollection import GXStandardObisCodeCollection
from .GXStandardObisCode import GXStandardObisCode
from .internal._GXCommon import _GXCommon
from .GXDate import GXDate
from .GXDateTime import GXDateTime
from .GXTime import GXTime
from .GXByteBuffer import GXByteBuffer
from .manufacturersettings.GXObisCode import GXObisCode

###Python 2 requires this
#pylint: disable=bad-option-value,old-style-class
class GXDLMSConverter:
    #
    # Constructor.
    #
    # value: Used standard.
    #
    def __init__(self, value=Standard.DLMS):
        self.standard = value
        # Collection of standard OBIS codes.
        self.codes = GXStandardObisCodeCollection()

    #
    # Get OBIS code description.
    #
    # @param logicalName
    # Logical name (OBIS code).
    # @param type
    # Object type.
    # @param description
    # Description filter.
    # Array of descriptions that match given OBIS code.
    #
    def getDescription(self, logicalName, type_=ObjectType.NONE, description=None):
        if not self.codes:
            self.__readStandardObisInfo(self.standard, self.codes)
        list_ = list()
        all_ = not logicalName
        for it in self.codes.find(logicalName, type_):
            if description and not it.description.lower().contains(description.lower()):
                continue
            if all_:
                list_.append("A=" + it.getOBIS()[0] + ", B=" + it.getOBIS()[1] + ", C=" + it.getOBIS()[2] + ", D=" + it.getOBIS()[3] + ", E=" + it.getOBIS()[4] + ", F=" + it.getOBIS()[5] + "\r\n" + it.description)
            else:
                list_.append(it.description)
        return list_

    #pylint: disable=too-many-boolean-expressions
    @classmethod
    def __updateOBISCodeInfo(cls, codes, it):
        if it.description:
            return
        ln = it.logicalName
        code_ = codes.find(ln, it.objectType)[0]
        if code_:
            it.description = code_.description
            if "10" in code_.dataType:
                code_.dataType = "10"
            elif "25" in code_.dataType or "26" in code_.dataType:
                code_.dataType = "25"
            elif "9" in code_.dataType:
                if (GXStandardObisCodeCollection.equalsMask2("0.0-64.96.7.10-14.255", ln) or GXStandardObisCodeCollection.equalsMask2("0.0-64.0.1.5.0-99,255", ln) or GXStandardObisCodeCollection.equalsMask2("0.0-64.0.1.2.0-99,255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.1.2.0-99,255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.1.5.0-99,255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.0.255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.6.255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.7.255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.13.255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.14.255", ln) or GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.15.255", ln)):
                    code_.dataType = "25"
                elif GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.1.255", ln):
                    code_.dataType = "27"
                elif GXStandardObisCodeCollection.equalsMask2("1.0-64.0.9.2.255", ln):
                    code_.dataType = "26"
            if not code_.dataType == "*" and not code_.dataType == "" and "," not in code_.dataType:
                type_ = code_.dataType
                it.uiDataType = 2, type_
        else:
            print("Unknown OBIS Code: " + it.logicalName + " Type: " + it.objectType)

    def updateOBISCodeInformation(self, objects):
        if not self.codes:
            self.__readStandardObisInfo(self.standard, self.codes)
        if isinstance(objects, list):
            for it in objects:
                self.__updateOBISCodeInfo(self.codes, it)
        else:
            self.__updateOBISCodeInfo(self.codes, objects)

    @classmethod
    def __getObjects(cls, standard):
        codes = list()
        if standard == Standard.ITALY:
            str_ = pkg_resources.resource_string(__name__, "Italy.txt").decode("utf-8")
        elif standard == Standard.INDIA:
            str_ = pkg_resources.resource_string(__name__, "India.txt").decode("utf-8")
        elif standard == Standard.SAUDI_ARABIA:
            str_ = pkg_resources.resource_string(__name__, "SaudiArabia.txt").decode("utf-8")
        if not str_:
            return None
        str_ = str_.replace("\r", " ")
        rows = str_.split('\n')
        for it in rows:
            items = it.split(';')
            if len(items) > 1:
                ot = ObjectType(int(items.get(0)))
                ln = _GXCommon.toLogicalName(_GXCommon.logicalNameToBytes(items.get(1)))
                version = int(items.get(2))
                desc = items.get(3)
                code_ = GXObisCode(ln, ot, desc)
                code_.version = version
                codes.append(code_)
        return codes

    @classmethod
    def __readStandardObisInfo(cls, standard, codes):
        if standard != Standard.DLMS:
            for it in cls.__getObjects(standard):
                tmp = GXStandardObisCode(None)
                tmp.interfaces = str(it.objectType)
                tmp.OBIS = it.logicalName.split('.')
                tmp.description = it.description
                codes.append(tmp)

        str_ = pkg_resources.resource_string(__name__, "OBISCodes.txt").decode("utf-8")
        str_ = str_.replace("\r", "")
        rows = str_.split('\n')
        for it in rows:
            if it != "":
                items = it.split(';')
                obis = items[0].split('.')
                try:
                    code_ = GXStandardObisCode(obis, str(items[3]) + "; " + str(items[4]) + "; " + str(items[5]) + "; " + str(items[6]) + "; " + str(items[7]), str(items[1]), str(items[2]))
                    codes.append(code_)
                except UnicodeEncodeError:
                    pass

    @classmethod
    def changeType(cls, value, type_):
        if _GXCommon.getDLMSDataType(value) == type_:
            return value
        if type_ == DataType.ARRAY:
            raise ValueError("Can't change array types.")
        if type_ == DataType.BCD:
            ret = int(value)
        elif type_ == DataType.BOOLEAN:
            ret = bool(value)
        elif type_ == DataType.COMPACT_ARRAY:
            raise ValueError("Can't change compact array types.")
        elif type_ == DataType.DATE:
            ret = GXDate(value)
        elif type_ == DataType.DATETIME:
            ret = GXDateTime(value)
        elif type_ == DataType.ENUM:
            raise ValueError("Can't change enumeration types.")
        elif type_ == DataType.FLOAT32:
            ret = float(value)
        elif type_ == DataType.FLOAT64:
            ret = float(value)
        elif type_ == DataType.INT16:
            ret = int(value)
        elif type_ == DataType.INT32:
            ret = int(value)
        elif type_ == DataType.INT64:
            ret = int(value)
        elif type_ == DataType.INT8:
            ret = int(value)
        elif type_ == DataType.NONE:
            ret = None
        elif type_ == DataType.OCTET_STRING:
            if isinstance(value, str):
                ret = GXByteBuffer.hexToBytes(str(value))
            else:
                raise ValueError("Can't change octect string type.")
        elif type_ == DataType.STRING:
            ret = str(value)
        elif type_ == DataType.BITSTRING:
            ret = str(value)
        elif type_ == DataType.STRING_UTF8:
            ret = str(value)
        elif type_ == DataType.STRUCTURE:
            raise ValueError("Can't change structure types.")
        elif type_ == DataType.TIME:
            ret = GXTime(value)
        elif type_ == DataType.UINT16:
            ret = int(value)
        elif type_ == DataType.UINT32:
            ret = int(value)
        elif type_ == DataType.UINT64:
            ret = int(value)
        elif type_ == DataType.UINT8:
            ret = int(value)
        else:
            raise ValueError('Invalid data type.')
        return ret



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#pylint: disable=broad-except,no-name-in-module
try:
    from enum import IntEnum
    __base = IntEnum
except Exception:
    __base = object

class ObjectType(__base):
    """ObjectType enumerates the usable types of DLMS objects in GuruxDLMS."""
    #pylint: disable=too-few-public-methods

    # When communicating with a meter, the application may demand periodical
    # actions.  If these actions are not linked to tariffication =
    # ActivityCalendar or Schedule, use an object of type ActionSchedule =
    # 0x16.
    ACTION_SCHEDULE = 22

    # When handling tariffication structures, you can use an object of type
    # ActivityCalendar.  It determines, when to activate specified scripts to
    # perform certain activities in the meter.  The activities, simply said,
    # scheduled actions, are operations that are carried out on a specified
    # day, at a specified time.  ActivityCalendar can be used together with a
    # more general object type, Schedule, and they can even overlap.  If
    # multiple actions are timed to the same moment, the actions determined in
    # the Schedule are executed first, and then the ones determined in the
    # ActivityCalendar.  If using object type SpecialDaysTable, with
    # ActivityCalendar, simultaneous actions determined in SpecialDaysTable are
    # executed over the ones determined in ActivityCalendar.  <p /> <b>Note:
    # </b>To make sure that tariffication is correct after a power failure,
    # only the latest missed action from ActivityCalendar is executed, with a
    # delay.  In a case of power failure, if a Schedule object coexists, the
    # latest missed action from ActivityCalendar has to be executed at the
    # correct time, sequentially with actions determined in the Schedule.
    ACTIVITY_CALENDAR = 20

    # All object types are used.
    NONE = 0

    # AssociationLogicalName object type is used with meters that utilize
    # Logical Name associations within a COSEM.
    ASSOCIATION_LOGICAL_NAME = 15

    # AssociationShortName object type is used with meters that utilize Short
    # Name associations within a COSEM.
    ASSOCIATION_SHORT_NAME = 12

    # To determine auto answering settings = for data transfer between device
    # and modem = s to answer incoming calls, use AutoAnswer object.
    AUTO_ANSWER = 28

    # To determine auto connecting settings = for data transfer from the meter
    # to defined destinations, use AutoConnect = previously known as AutoDial
    # object.
    AUTO_CONNECT = 29

    # An object of type Clock is used to handle the information of a date =
    # day, month and year and/or a time = hundredths of a second, seconds,
    # minutes and hours.
    CLOCK = 8

    # An object of type Data typically stores manufacturer specific information
    # of the meter, for example configuration data and logical name.
    DATA = 1

    # An object of type DemandRegister stores a value, information of the item,
    # which the value belongs to, the status of the item, and the time of the
    # value.  DemandRegister object type enables both current, and last
    # average,
    # it supports both block, and sliding demand calculation, and it also
    # provides resetting the value average, and periodic averages.
    DEMAND_REGISTER = 5

    # MAC address of the physical device.  The name and the use of this
    # interface class has been changed from Ethernet setup to MAC address setup
    # to allow a more general use.
    MAC_ADDRESS_SETUP = 43


    # ExtendedRegister stores a value, and understands the type of the value.
    # Refer to an object of this type by its logical name, using the OBIS
    # identification code.
    EXTENDED_REGISTER = 4

    # To determine the GPRS settings, use GprsSetup object.
    GPRS_SETUP = 45

    # To determine the HDLC = High-level Data Link Control settings, use the
    # IecHdlcSetup object.
    IEC_HDLC_SETUP = 23

    # To determine the Local Port settings, use the IecLocalPortSetup object.
    IEC_LOCAL_PORT_SETUP = 19

    # To determine the Twisted Pair settings, use the IecTwistedPairSetup
    # object.
    IEC_TWISTED_PAIR_SETUP = 24

    # To determine the IP 4 settings, use the Ip4Setup object.
    IP4_SETUP = 42

    GSM_DIAGNOSTIC = 47

    # To determine the IP 6 settings, use the Ip6Setup object.
    IP6_SETUP = 48

    # To determine the M-BUS settings, use the MbusSetup object.
    MBUS_SLAVE_PORT_SETUP = 25

    # To determine modem settings, use ModemConfiguration object.
    MODEM_CONFIGURATION = 27

    # To determine PPP = Point-to-Point Protocol settings, use the PppSetup
    # object.
    PPP_SETUP = 44

    # ProfileGeneric determines a general way of gathering values from a
    # profile.  The data is retrieved either by a period of time, or by an
    # occuring event.  When gathering values from a profile, you need to
    # understand the concept of the profile buffer, in which the profile data
    # is stored.  The buffer may be sorted by a register, or by a clock, within
    # the profile, or the data can be just piled in it, in order: last in,
    # first out.  You can retrieve a part of the buffer, within a certain range
    # of values, or by a range of entry numbers.  You can also determine
    # objects, whose values are to be retained.  To determine, what to
    # retrieve,
    # and what to retain, you need to assign the objects to the profile.  You
    # can use static assignments, as all entries in a buffer are alike = same
    # size, same structure.  <p /> <b>Note: </b>When you modify any assignment,
    # the buffer of the corresponding profile is cleared, and all other
    # profiles, using the modified one, will be cleared too.  This is to make
    # sure that their entries stay alike by size and structure.
    PROFILE_GENERIC = 7

    # Register stores a value, and understands the type of the value.  Refer to
    # an object of this type by its logical name, using the OBIS identification
    # code.
    REGISTER = 3

    # When handling tariffication structures, you can use RegisterActivation to
    # determine, what objects to enable, when activating a certain activation
    # mask.  The objects, assigned to the register, but not determined in the
    # mask, are disabled.  <p /> <b>Note: </b>If an object is not assigned to
    # any register, it is, by default, enabled.
    REGISTER_ACTIVATION = 6

    # RegisterMonitor allows you to determine scripts to execute, when a
    # register value crosses a specified threshold.  To use RegisterMonitor,
    # also ScriptTable needs to be instantiated in the same logical device.
    REGISTER_MONITOR = 21

    # Instances of the Disconnect control IC manage an internal or external
    # disconnect unit of the meter.
    DISCONNECT_CONTROL = 70

    LIMITER = 71

    MBUS_CLIENT = 72

    PUSH_SETUP = 40

    COMPACT_DATA = 62

    PARAMETER_MONITOR = 65
    WIRELESS_MODE_Q_CHANNEL = 73
    MBUS_MASTER_PORT_SETUP = 74

    # RegisterTable stores identical attributes of objects, in a selected
    # collection of objects.  All the objects in the collection need to be of
    # the same type.  Also, the value in value groups A to D and F in their
    # logical name = OBIS identification code needs to be identical.  <p />
    # Clause 5 determines the possible values in value group E, as a table,
    # where header = the common part, and each cell = a possible E value, of
    # the OBIS code.
    REGISTER_TABLE = 61

    #
    # Configure a ZigBee PRO device with information necessary to create or
    # join the network.
    ZIG_BEE_SAS_STARTUP = 101

    #
    # Configure the behavior of a ZigBee PRO device on joining or loss of
    # connection to the network.
    ZIG_BEE_SAS_JOIN = 102

    #
    # Configure the fragmentation feature of ZigBee PRO transport layer.
    ZIG_BEE_SAS_APS_FRAGMENTATION = 103
    ZIG_BEE_NETWORK_CONTROL = 104

    DATA_PROTECTION = 30

    ACCOUNT = 111

    CREDIT = 112

    CHARGE = 113

    TOKEN_GATEWAY = 115

    # SapAssigment stores information of assignment of the logical devices to
    # their SAP = Service Access Points.
    SAP_ASSIGNMENT = 17

    # Instances of the Image transfer IC model the mechanism of transferring
    # binary files, called firmware Images to COSEM servers.
    IMAGE_TRANSFER = 18

    # To handle time and date driven actions, use Schedule, with an object of
    # type SpecialDaysTable.
    SCHEDULE = 10

    # To trigger a set of actions with an execute method, use object type
    # ScriptTable.  Each table entry = script includes a unique identifier, and
    # a set of action specifications, which either execute a method, or modify
    # the object attributes, within the logical device.  The script can be
    # triggered by other objects = within the same logical device, or from the
    # outside.
    SCRIPT_TABLE = 9

    # To determine the SMTP protocol settings, use the SmtpSetup object.
    SMTP_SETUP = 2

    # With SpecialDaysTable you can determine dates to override a preset
    # behaviour, for specific days = data item day_id.  SpecialDaysTable works
    # together with objects of Schedule, or Activity Calendar.
    SPECIAL_DAYS_TABLE = 11

    # StatusMapping object stores status words with mapping.  Each bit in the
    # status word is mapped to position = s in referencing status table.
    STATUS_MAPPING = 63

    SECURITY_SETUP = 64

    # To determine Internet TCP/UDP protocol settings, use the TcpUdpSetup
    # object.
    TCP_UDP_SETUP = 41

    # In an object of type UtilityTables each "Table" = ANSI C12.19:1997 table
    # data is represented as an instance, and identified by its logical name.
    UTILITY_TABLES = 26

    #
    # S-FSK Phy MAC Setup
    SFSK_PHY_MAC_SETUP = 50


    # S-FSK Active initiator.
    SFSK_ACTIVE_INITIATOR = 51

    # S-FSK MAC synchronization timeouts
    SFSK_MAC_SYNCHRONIZATION_TIMEOUTS = 52

    # S-FSK MAC Counters.
    SFSK_MAC_COUNTERS = 53

    # G3-PLC MAC layer counters
    G3_PLC_MAC_LAYER_COUNTERS = 90


    # G3-PLC MAC setup.
    G3_PLC_MAC_SETUP = 91

    # G3-PLC 6LoWPAN.
    G3_PLC6_LO_WPAN = 92

    #
    # Tariff Plan =Piano Tariffario) is used in Italian standard UNI/TS
    # 11291-11.
    TARIFF_PLAN = 8192



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums import ObjectType
#pylint: disable=bad-option-value,too-many-locals, cyclic-import,old-style-class,too-few-public-methods
class _GXObjectFactory:
    #Reserved for internal use.

    #
    # Constructor.
    def __init__(self):
        pass

    @classmethod
    def createObject(cls, ot):
        #pylint: disable=bad-option-value,redefined-variable-type
        from .objects import GXDLMSObject, GXDLMSActionSchedule, GXDLMSActivityCalendar, GXDLMSAssociationLogicalName,\
            GXDLMSAssociationShortName, GXDLMSAutoAnswer, GXDLMSAutoConnect, GXDLMSClock, GXDLMSData, GXDLMSDemandRegister,\
            GXDLMSMacAddressSetup, GXDLMSRegister, GXDLMSExtendedRegister, GXDLMSGprsSetup, GXDLMSHdlcSetup, GXDLMSIECOpticalPortSetup,\
            GXDLMSIecTwistedPairSetup, GXDLMSIp4Setup, GXDLMSIp6Setup, GXDLMSMBusSlavePortSetup, GXDLMSImageTransfer, GXDLMSSecuritySetup,\
            GXDLMSDisconnectControl, GXDLMSLimiter, GXDLMSMBusClient, GXDLMSModemConfiguration, GXDLMSPppSetup, GXDLMSProfileGeneric,\
            GXDLMSRegisterMonitor, GXDLMSRegisterActivation, GXDLMSSapAssignment, GXDLMSSchedule, GXDLMSScriptTable, GXDLMSSpecialDaysTable,\
            GXDLMSTcpUdpSetup, GXDLMSPushSetup, GXDLMSMBusMasterPortSetup, GXDLMSGSMDiagnostic, GXDLMSAccount, GXDLMSCredit, GXDLMSCharge, \
            GXDLMSTokenGateway, GXDLMSParameterMonitor
        #  If IC is manufacturer specific or unknown.
        if ot is None:
            raise ValueError("Invalid object type.")

        if ot == ObjectType.ACTION_SCHEDULE:
            ret = GXDLMSActionSchedule()
        elif ot == ObjectType.ACTIVITY_CALENDAR:
            ret = GXDLMSActivityCalendar()
        elif ot == ObjectType.ASSOCIATION_LOGICAL_NAME:
            ret = GXDLMSAssociationLogicalName()
        elif ot == ObjectType.ASSOCIATION_SHORT_NAME:
            ret = GXDLMSAssociationShortName()
        elif ot == ObjectType.AUTO_ANSWER:
            ret = GXDLMSAutoAnswer()
        elif ot == ObjectType.AUTO_CONNECT:
            ret = GXDLMSAutoConnect()
        elif ot == ObjectType.CLOCK:
            ret = GXDLMSClock()
        elif ot == ObjectType.DATA:
            ret = GXDLMSData()
        elif ot == ObjectType.DEMAND_REGISTER:
            ret = GXDLMSDemandRegister()
        elif ot == ObjectType.MAC_ADDRESS_SETUP:
            ret = GXDLMSMacAddressSetup()
        elif ot == ObjectType.REGISTER:
            ret = GXDLMSRegister()
        elif ot == ObjectType.EXTENDED_REGISTER:
            ret = GXDLMSExtendedRegister()
        elif ot == ObjectType.GPRS_SETUP:
            ret = GXDLMSGprsSetup()
        elif ot == ObjectType.IEC_HDLC_SETUP:
            ret = GXDLMSHdlcSetup()
        elif ot == ObjectType.IEC_LOCAL_PORT_SETUP:
            ret = GXDLMSIECOpticalPortSetup()
        elif ot == ObjectType.IEC_TWISTED_PAIR_SETUP:
            ret = GXDLMSIecTwistedPairSetup()
        elif ot == ObjectType.IP4_SETUP:
            ret = GXDLMSIp4Setup()
        elif ot == ObjectType.IP6_SETUP:
            ret = GXDLMSIp6Setup()
        elif ot == ObjectType.MBUS_SLAVE_PORT_SETUP:
            ret = GXDLMSMBusSlavePortSetup()
        elif ot == ObjectType.IMAGE_TRANSFER:
            ret = GXDLMSImageTransfer()
        elif ot == ObjectType.SECURITY_SETUP:
            ret = GXDLMSSecuritySetup()
        elif ot == ObjectType.DISCONNECT_CONTROL:
            ret = GXDLMSDisconnectControl()
        elif ot == ObjectType.LIMITER:
            ret = GXDLMSLimiter()
        elif ot == ObjectType.MBUS_CLIENT:
            ret = GXDLMSMBusClient()
        elif ot == ObjectType.MODEM_CONFIGURATION:
            ret = GXDLMSModemConfiguration()
        elif ot == ObjectType.PPP_SETUP:
            ret = GXDLMSPppSetup()
        elif ot == ObjectType.PROFILE_GENERIC:
            ret = GXDLMSProfileGeneric()
        elif ot == ObjectType.REGISTER_MONITOR:
            ret = GXDLMSRegisterMonitor()
        elif ot == ObjectType.REGISTER_ACTIVATION:
            ret = GXDLMSRegisterActivation()
        elif ot == ObjectType.REGISTER_TABLE:
            ret = GXDLMSObject(ot)
        elif ot == ObjectType.ZIG_BEE_SAS_STARTUP:
            ret = GXDLMSObject(ot)
        elif ot == ObjectType.ZIG_BEE_SAS_JOIN:
            ret = GXDLMSObject(ot)
        elif ot == ObjectType.SAP_ASSIGNMENT:
            ret = GXDLMSSapAssignment()
        elif ot == ObjectType.SCHEDULE:
            ret = GXDLMSSchedule()
        elif ot == ObjectType.SCRIPT_TABLE:
            ret = GXDLMSScriptTable()
        elif ot == ObjectType.SPECIAL_DAYS_TABLE:
            ret = GXDLMSSpecialDaysTable()
        elif ot == ObjectType.STATUS_MAPPING:
            ret = GXDLMSObject(ot)
        elif ot == ObjectType.TCP_UDP_SETUP:
            ret = GXDLMSTcpUdpSetup()
        elif ot == ObjectType.ZIG_BEE_SAS_APS_FRAGMENTATION:
            ret = GXDLMSObject(ot)
        elif ot == ObjectType.UTILITY_TABLES:
            ret = GXDLMSObject(ot)
        elif ot == ObjectType.PUSH_SETUP:
            ret = GXDLMSPushSetup()
        elif ot == ObjectType.MBUS_MASTER_PORT_SETUP:
            ret = GXDLMSMBusMasterPortSetup()
        elif ot == ObjectType.GSM_DIAGNOSTIC:
            ret = GXDLMSGSMDiagnostic()
        elif ot == ObjectType.ACCOUNT:
            ret = GXDLMSAccount()
        elif ot == ObjectType.CREDIT:
            ret = GXDLMSCredit()
        elif ot == ObjectType.CHARGE:
            ret = GXDLMSCharge()
        elif ot == ObjectType.TOKEN_GATEWAY:
            ret = GXDLMSTokenGateway()
        elif ot == ObjectType.PARAMETER_MONITOR:
            ret = GXDLMSParameterMonitor()
        else:
            ret = GXDLMSObject(ot)
        return ret



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode, ObjectType, DataType
from ..internal._GXCommon import _GXCommon
from .enums import BaudRate

# pylint: disable=too-many-instance-attributes
class GXDLMSHdlcSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSHdlcSetup
    """

    def __init__(self, ln="0.0.22.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.IEC_HDLC_SETUP, ln, sn)
        self.communicationSpeed = BaudRate.BAUDRATE_9600
        self.windowSizeTransmit = 1
        self.windowSizeReceive = 1
        self.maximumInfoLengthReceive = 128
        self.maximumInfoLengthTransmit = 128
        self.inactivityTimeout = 120
        self.version = 1
        self.interCharachterTimeout = 0
        self.deviceAddress = 0

    def getValues(self):
        return [self.logicalName,
                self.communicationSpeed,
                self.windowSizeTransmit,
                self.windowSizeReceive,
                self.maximumInfoLengthTransmit,
                self.maximumInfoLengthReceive,
                self.interCharachterTimeout,
                self.inactivityTimeout,
                self.deviceAddress]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Communication speed
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  Window size transmit
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  Window size receive
        if all_ or not self.isRead(4):
            attributes.append(4)
        #  maximum info length transmit
        if all_ or not self.isRead(5):
            attributes.append(5)
        #  MaximumInfoLengthReceive
        if all_ or not self.isRead(6):
            attributes.append(6)
        #  InterCharachterTimeout
        if all_ or not self.isRead(7):
            attributes.append(7)
        #  Inactivity timeout
        if all_ or not self.isRead(8):
            attributes.append(8)
        #  Device address
        if all_ or not self.isRead(9):
            attributes.append(9)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 9

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ENUM
        elif index == 3:
            ret = DataType.UINT8
        elif index == 4:
            ret = DataType.UINT8
        elif index == 5:
            if self.version == 0:
                ret = DataType.UINT8
            else:
                ret = DataType.UINT16
        elif index == 6:
            if self.version == 0:
                ret = DataType.UINT8
            else:
                ret = DataType.UINT16
        elif index == 7:
            ret = DataType.UINT16
        elif index == 8:
            ret = DataType.UINT16
        elif index == 9:
            ret = DataType.UINT16
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.communicationSpeed
        elif e.index == 3:
            ret = self.windowSizeTransmit
        elif e.index == 4:
            ret = self.windowSizeReceive
        elif e.index == 5:
            ret = self.maximumInfoLengthTransmit
        elif e.index == 6:
            ret = self.maximumInfoLengthReceive
        elif e.index == 7:
            ret = self.interCharachterTimeout
        elif e.index == 8:
            ret = self.inactivityTimeout
        elif e.index == 9:
            ret = self.deviceAddress
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.communicationSpeed = e.value
        elif e.index == 3:
            self.windowSizeTransmit = e.value
        elif e.index == 4:
            self.windowSizeReceive = e.value
        elif e.index == 5:
            self.maximumInfoLengthTransmit = e.value
        elif e.index == 6:
            self.maximumInfoLengthReceive = e.value
        elif e.index == 7:
            self.interCharachterTimeout = e.value
        elif e.index == 8:
            self.inactivityTimeout = e.value
        elif e.index == 9:
            self.deviceAddress = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.communicationSpeed = BaudRate(reader.readElementContentAsInt("Speed"))
        self.windowSizeTransmit = reader.readElementContentAsInt("WindowSizeTx")
        self.windowSizeReceive = reader.readElementContentAsInt("WindowSizeRx")
        self.maximumInfoLengthTransmit = reader.readElementContentAsInt("MaximumInfoLengthTx")
        self.maximumInfoLengthReceive = reader.readElementContentAsInt("MaximumInfoLengthRx")
        self.interCharachterTimeout = reader.readElementContentAsInt("InterCharachterTimeout")
        self.inactivityTimeout = reader.readElementContentAsInt("InactivityTimeout")
        self.deviceAddress = reader.readElementContentAsInt("DeviceAddress")

    def save(self, writer):
        writer.writeElementString("Speed", self.communicationSpeed)
        writer.writeElementString("WindowSizeTx", self.windowSizeTransmit)
        writer.writeElementString("WindowSizeRx", self.windowSizeReceive)
        writer.writeElementString("MaximumInfoLengthTx", self.maximumInfoLengthTransmit)
        writer.writeElementString("MaximumInfoLengthRx", self.maximumInfoLengthReceive)
        writer.writeElementString("InterCharachterTimeout", self.interCharachterTimeout)
        writer.writeElementString("InactivityTimeout", self.inactivityTimeout)
        writer.writeElementString("DeviceAddress", self.deviceAddress)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums import DataType

# pylint: disable=bad-option-value,old-style-class,too-few-public-methods
class GXWriteItem:
    #
    # Constructor.
    #
    # @param object
    #            Object to write.
    # @param attributeIndex
    #            Attribute index.
    #
    def __init__(self, object_=None, attributeIndex=0):
        self.dataType = DataType.NONE
        self.target = object_
        # Attribute index to write.
        self.index = attributeIndex
        # Written object.
        self.target = None
        # Data type to write.
        self.dataType = None
        # Parameter selector.
        self.selector = 0
        # Optional parameters.
        self.parameters = None



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#pylint: disable=broad-except,no-name-in-module
try:
    from enum import IntEnum
    __base = IntEnum
except Exception:
    __base = object

class BerType(__base):
    """
    BER encoding enumeration values.
    """
    #pylint: disable=too-few-public-methods

    # End of Content.
    EOC = 0x00

    # Boolean.
    BOOLEAN = 0x1

    # Integer.
    INTEGER = 0x2

    # Bit String.
    BIT_STRING = 0x3

    # Octet string.
    OCTET_STRING = 0x4

    # Null value.
    NULL = 0x5

    # Object identifier.
    OBJECT_IDENTIFIER = 0x6

    # Object Descriptor.
    OBJECT_DESCRIPTOR = 7

    # External
    EXTERNAL = 8

    # Real = float).
    REAL = 9

    # Enumerated.
    ENUMERATED = 10

    # Sequence.
    SEQUENCE = 0x10

    SET = 0x11

    # Utf8 String.
    UTF8STRING = 12

    # Numeric string.
    NUMERIC_STRING = 18

    # Printable string.
    PRINTABLE_STRING = 19

    # Teletex string.
    TELETEX_STRING = 20

    # Videotex string.
    VIDEOTEX_STRING = 21

    # Ia5 string
    IA5_STRING = 22

    # Utc time.
    UTC_TIME = 23

    # Generalized time.
    GENERALIZED_TIME = 24

    # Graphic string.
    GRAPHIC_STRING = 25

    # Visible string.
    VISIBLE_STRING = 26

    # General string.
    GENERAL_STRING = 27

    # Universal string.
    UNIVERSAL_STRING = 28

    # Bmp string.
    BMP_STRING = 30

    # Application class.
    APPLICATION = 0x40

    # Context class.
    CONTEXT = 0x80

    # Private class.
    PRIVATE = 0xc0

    # Constructed.
    CONSTRUCTED = 0x20



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#pylint: disable=broad-except,no-name-in-module
try:
    from enum import IntEnum
    __base = IntEnum
except Exception:
    __base = object

class Authentication(__base):
    """
    Authentication enumerates the authentication levels.
    """
    #pylint: disable=too-few-public-methods

    #No authentication is used.
    NONE = 0

    #Low authentication is used.
    LOW = 1

    #High authentication is used.
    HIGH = 2

    #High authentication is used.  Password is hashed with MD5.
    HIGH_MD5 = 3

    #High authentication is used.  Password is hashed with SHA1.
    HIGH_SHA1 = 4

    #High authentication is used.  Password is hashed with GMAC.
    HIGH_GMAC = 5

    #High authentication is used.  Password is hashed with SHA-256.
    HIGH_SHA256 = 6

    #High authentication is used.  Password is hashed with ECDSA.
    HIGH_ECDSA = 7

    @classmethod
    def valueofString(cls, value):
        return Authentication[value.upper()]



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType, AccessMode, MethodAccessMode, Authentication
from .GXDLMSObjectCollection import GXDLMSObjectCollection
from ..GXSecure import GXSecure
from ..ConnectionState import ConnectionState

# pylint: disable=too-many-instance-attributes
class GXDLMSAssociationShortName(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSAssociationShortName
    """

    def __init__(self, ln="0.0.40.0.0.255", sn=0xFA00):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.ASSOCIATION_SHORT_NAME, ln, sn)
        self.secret = bytearray()
        self.objectList = GXDLMSObjectCollection(self)
        self.version = 2
        self.securitySetupReference = None

    def getValues(self):
        return [self.logicalName,
                self.objectList,
                None,
                self.securitySetupReference]

    #
    # Invokes method.
    # @param index Method index.
    #
    def invoke(self, settings, e):
        #  Check reply_to_HLS_authentication
        if e.index == 8:
            serverChallenge = None
            clientChallenge = None
            ic = 0
            readSecret = []
            accept = False
            if settings.authentication == Authentication.HIGH_ECDSA:
                accept = False
            else:
                if settings.authentication == Authentication.HIGH_GMAC:
                    readSecret = settings.sourceSystemTitle
                    bb = GXByteBuffer(e.parameters)
                    bb.getUInt8()
                    ic = bb.getUInt32()
                else:
                    readSecret = self.secret
                serverChallenge = GXSecure.secure(settings, settings.cipher, ic, settings.getStoCChallenge(), readSecret)
                clientChallenge = int(e.parameters)
                accept = serverChallenge == clientChallenge
            if accept:
                if settings.authentication == Authentication.HIGH_GMAC:
                    readSecret = settings.cipher.getSystemTitle()
                    ic = settings.cipher.invocationCounter
                else:
                    readSecret = self.secret
                settings.setConnected(settings.connected | ConnectionState.DLMS)
                return GXSecure.secure(settings, settings.cipher, ic, settings.getCtoSChallenge(), readSecret)
        else:
            settings.setConnected(settings.connected & ~ConnectionState.DLMS)
            e.error = ErrorCode.READ_WRITE_DENIED
        return None

    def getAttributeIndexToRead(self, all_):
        attributes = list()
        if all_ or not self.logicalName:
            attributes.append(1)
        if all_ or not self.isRead(2):
            attributes.append(2)
        if self.version > 1:
            if all_ or not self.isRead(3):
                attributes.append(3)
            if all_ or not self.isRead(4):
                attributes.append(4)
        return attributes

    def getAttributeCount(self):
        if self.version < 2:
            return 2
        return 4

    def getMethodCount(self):
        return 8

    @classmethod
    def __getAccessRights_(cls, item, data):
        data.setUInt8(DataType.STRUCTURE)
        data.setUInt8(3)
        _GXCommon.setData(data, DataType.UINT16, item.shortName)
        data.setUInt8(int(DataType.ARRAY))
        data.setUInt8(len(item.attributes))
        for att in item.attributes:
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(3)
            _GXCommon.setData(data, DataType.INT8, att.index)
            _GXCommon.setData(data, DataType.ENUM, att.access.value)
            _GXCommon.setData(data, DataType.NONE, None)
        data.setUInt8(DataType.ARRAY)
        data.setUInt8(len(item.methodAttributes))
        for it in item.methodAttributes:
            data.setUInt8(DataType.STRUCTURE)
            data.setUInt8(2)
            _GXCommon.setData(data, DataType.INT8, it.index)
            _GXCommon.setData(data, DataType.ENUM, it.getMethodAccess().value)

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ARRAY
        elif index == 3:
            ret = DataType.ARRAY
        elif index == 4:
            return DataType.OCTET_STRING
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    def getObjects(self, settings, e):
        bb = GXByteBuffer()
        cnt = len(self.objectList)
        if settings.index == 0:
            settings.setCount(cnt)
            bb.setUInt8(int(DataType.ARRAY))
            _GXCommon.setObjectCount(cnt, bb)
        pos = 0
        if cnt != 0:
            for it in self.objectList:
                pos += 1
                if not pos <= settings.index:
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(4)
                    _GXCommon.setData(bb, DataType.INT16, it.shortName)
                    _GXCommon.setData(bb, DataType.UINT16, it.objectType)
                    _GXCommon.setData(bb, DataType.UINT8, 0)
                    _GXCommon.setData(bb, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(it.logicalName))
                    settings.index = settings.index + 1
                    if settings.isServer:
                        if not e.isSkipMaxPduSize() and len(bb) >= settings.maxPduSize:
                            break
        return bb.array()

    def getValue(self, settings, e):
        ret = None
        bb = GXByteBuffer()
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.getObjects(settings, e)
        elif e.index == 3:
            lnExists = self.objectList.findBySN(self.shortName) is not None
            cnt = len(self.objectList)
            if not lnExists:
                cnt += 1
            bb.setUInt8(DataType.ARRAY)
            _GXCommon.setObjectCount(cnt, bb)
            for it in self.objectList:
                self.__getAccessRights_(it, bb)
            if not lnExists:
                self.__getAccessRights_(self, bb)
            ret = bb.array()
        elif e.index == 4:
            ret = _GXCommon.getBytes(self.securitySetupReference)
        e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    def updateAccessRights(self, buff):
        for access in buff:
            sn = access[0]
            obj = self.objectList.findBySN(sn)
            if obj:
                for attributeAccess in access[1]:
                    id1 = attributeAccess[0]
                    mode1 = AccessMode(attributeAccess[1])
                    obj.setAccess(id1, mode1)
                for methodAccess in access[2]:
                    id2 = methodAccess[0]
                    mode2 = MethodAccessMode(methodAccess[1])
                    obj.setMethodAccess(id2, mode2)

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            from .._GXObjectFactory import _GXObjectFactory
            self.objectList = []
            if e.value:
                for item in e.value:
                    sn = item[0]
                    ot = ObjectType(item[1])
                    version = item[2]
                    ln = _GXCommon.toLogicalName(item[3])
                    obj = _GXObjectFactory.createObject(ot)
                    obj.logicalName = ln
                    obj.shortName = sn
                    obj.version = version
                    self.objectList.append(obj)
        elif e.index == 3:
            if e.value is None:
                for it in self.objectList:
                    pos = 1
                    while pos != it.getAttributeCount():
                        it.setAccess(pos, AccessMode.NO_ACCESS)
                        pos += 1
            else:
                self.updateAccessRights(e.value)
        elif e.index == 4:
            self.securitySetupReference = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        str_ = reader.readElementContentAsString("Secret")
        if str_ is None:
            self.secret = None
        else:
            self.secret = GXByteBuffer.hexToBytes(str_)
        self.securitySetupReference = reader.readElementContentAsString("SecuritySetupReference")

    def save(self, writer):
        writer.writeElementString("Secret", GXByteBuffer.hex(self.secret))
        writer.writeElementString("SecuritySetupReference", self.securitySetupReference)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSObjectDefinition import GXDLMSObjectDefinition

# pylint: disable=too-many-instance-attributes
class GXDLMSRegisterActivation(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSRegisterActivation
    """

    #
    # Constructor.
    #
    # @param ln
    #            Logical Name of the object.
    # @param sn
    #            Short Name of the object.
    #
    def __init__(self, ln=None, sn=0):
        super(GXDLMSRegisterActivation, self).__init__(ObjectType.REGISTER_ACTIVATION, ln, sn)
        self.registerAssignment = list()
        self.maskList = list()
        self.activeMask = bytearray()

    def getValues(self):
        return [self.logicalName,
                self.registerAssignment,
                self.maskList,
                self.activeMask]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  MaskList
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  ActiveMask
        if all_ or not self.isRead(4):
            attributes.append(4)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 4

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 3

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ARRAY
        elif index == 3:
            ret = DataType.ARRAY
        elif index == 4:
            ret = DataType.OCTET_STRING
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            if not self.registerAssignment:
                data.setUInt8(0)
            else:
                data.setUInt8(len(self.registerAssignment))
                for it in self.registerAssignment:
                    data.setUInt8(DataType.STRUCTURE)
                    data.setUInt8(2)
                    _GXCommon.setData(data, DataType.UINT16, it.objectType)
                    _GXCommon.setData(data, DataType.OCTET_STRING, _GXCommon.logicalNameToBytes(it.logicalName))
            return data
        if e.index == 3:
            data = GXByteBuffer()
            data.setUInt8(DataType.ARRAY)
            data.setUInt8(len(self.maskList))
            for k, v in self.maskList:
                data.setUInt8(DataType.STRUCTURE)
                data.setUInt8(2)
                _GXCommon.setData(data, DataType.OCTET_STRING, k)
                data.setUInt8(DataType.ARRAY)
                data.setUInt8(len(v))
                for b in v:
                    _GXCommon.setData(data, DataType.UINT8, b)
            return data
        if e.index == 4:
            return self.activeMask
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.registerAssignment = []
            if e.value:
                for it in e.value:
                    item = GXDLMSObjectDefinition()
                    item.objectType = ObjectType(it[0])
                    item.logicalName = _GXCommon.toLogicalName(it[1])
                    self.registerAssignment.append(item)
        elif e.index == 3:
            self.maskList = []
            if e.value:
                for it in e.value:
                    self.maskList.append((it[0], it[0]))
        elif e.index == 4:
            if not e.value:
                self.activeMask = None
            else:
                self.activeMask = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.registerAssignment = []
        if reader.isStartElement("RegisterAssignment", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSObjectDefinition()
                it.objectType = ObjectType(reader.readElementContentAsInt("ObjectType"))
                it.logicalName = reader.readElementContentAsString("LN")
                self.registerAssignment.append(it)
        self.maskList = []
        if reader.isStartElement("MaskList", True):
            while reader.isStartElement("Item", True):
                mask = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Mask"))
                str_ = reader.readElementContentAsString("Index")
                i = GXByteBuffer.hexToBytes(str_.replace(";", " "))
                self.maskList.append((mask, i))
            reader.readEndElement("MaskList")
        self.activeMask = GXByteBuffer.hexToBytes(reader.readElementContentAsString("ActiveMask"))

    def save(self, writer):
        if self.registerAssignment:
            writer.writeStartElement("RegisterAssignment")
            for it in self.registerAssignment:
                writer.writeStartElement("Item")
                writer.writeElementString("ObjectType", it.objectType)
                writer.writeElementString("LN", it.logicalName)
                writer.writeEndElement()
            writer.writeEndElement()

        if self.maskList:
            writer.writeStartElement("MaskList")
            for k, v in self.maskList:
                writer.writeStartElement("Item")
                writer.writeElementString("Mask", GXByteBuffer.hex(k))
                writer.writeElementString("Index", GXByteBuffer.hex(v).replace(" ", ";"))
                writer.writeEndElement()
            writer.writeEndElement()
        if self.activeMask:
            writer.writeElementString("ActiveMask", GXByteBuffer.hex(self.activeMask))
